import os
import socket
import threading
import time

from .utils import Stream, Verify, File
from .database import DataBaseManager


class Base:
    def __init__(self, **kwargs):
        self.service_host = kwargs.get('service_host', '127.0.0.1')
        self.service_port = kwargs.get('service_port', 5000)
        self.encoding = kwargs.get('encoding', 'utf8')
        self.buffer_size = kwargs.get('buffer_size', 4096)
        self.stream = Stream()
        self.verify = Verify()
        self.file = File()

    def __str__(self):
        return str(self.__dict__)

    def current(self, order: str) -> bool:
        """ 通用服务 """
        if order == 'help':
            self.stream.system_out_help()
        elif order == 'me':
            self.stream.system_out_normal(f'{self.username} {self.address}')
        elif order == 'clear':
            os.system('cls')
            self.stream.start_client()
        elif order == 'path':
            self.stream.system_out_path()

        return order in ['help', 'me', 'clear', 'path']


# ======================================================================================================================
class Service(Base):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.service = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.max_client = kwargs.get('max_client', 5)
        self.client_pool = dict()
        self.data = DataBaseManager(**kwargs)
        self.lock = threading.Lock()

    def send_all(self, msg: str):
        # 给所有用户发送消息
        for client in self.client_pool.values():
            client.send(msg.encode(self.encoding))

    def send_user(self, user: str, to_user: str, msg: str) -> bool:
        # 给指定用户发送消息
        if to_user in self.client_pool.keys():
            self.client_pool[to_user].send(f'{user}: {msg}'.encode(self.encoding))
            return True

        elif to_user == 'all':
            self.send_all(f'[all] {user}: {msg}')
            return True

    def login_or_register(self, order: str, username: str, password: str) -> str:
        result = 'ok'
        if order == 'login' and (not self.data.login(username, password)):
            result = 'err 用户名或密码错误'
        elif order == 'register' and (not self.data.register(username, password)):
            result = 'err 用户名已经存在, 注册失败'
        elif username in self.client_pool.keys():
            result = 'err 用户已经在线!'
        return result

    def main(self, client: socket.socket, address: tuple, username: str):
        self.client_pool[username] = client
        while True:
            try:
                try:
                    order = client.recv(self.buffer_size).decode(self.encoding)
                except Exception as e:
                    continue

                if order == 'exit':
                    # 断开连接
                    client.send('exit'.encode(self.encoding))
                    client.close()
                    self.client_pool.pop(username)
                    self.stream.system_out_warning(f'client: {username}{address} 离开聊天室')
                    self.send_all(f'用户 {username} 离开聊天室')
                    break

                elif order == 'ls':
                    # 返回所有在线用户
                    msg = ', '.join(self.client_pool.keys())
                    client.send(f'users: [{msg}]'.encode(self.encoding))

                elif order.startswith('chat'):
                    # 给指定用户发送消息
                    order = order.split(' ')
                    user = order[1]
                    msg = ' '.join(order[2:])
                    if not self.send_user(user=username, to_user=user, msg=msg):
                        # 发送失败
                        client.send(f'用户: {user}, 不存在'.encode(self.encoding))

                elif order.startswith('file'):
                    # 文件传输
                    self.lock.acquire()
                    order, filename, filesize, to, user = order.split(' ')
                    filename = os.path.basename(filename)
                    data = self.file.download_file(client=client, filesize=int(filesize), filename=filename)

                    if user in self.client_pool.keys():
                        self.stream.system_out_normal(f'文件 {filename} 将转发给: {user}')
                        self.client_pool[user].send(f'file {filename} {filesize} {username}'.encode(self.encoding))
                        self.file.upload_file(filename=filename, filesize=int(filesize), client=self.client_pool[user])

                    else:
                        # 没有此人
                        client.send(f'用户: {user} 不存在'.encode(self.encoding))
                    self.lock.release()

                elif order.startswith('repwd'):
                    # 修改密码
                    order, old_pwd, new_pwd = order.split(' ')
                    flag = self.data.re_pwd(username=username, old_pwd=old_pwd, new_pwd=new_pwd)
                    # 修改失败, 返回提示
                    if not flag:
                        client.send(f'密码不匹配, 修改失败'.encode(self.encoding))
                    else:
                        client.send(f'密码修改成功'.encode(self.encoding))

                else:
                    client.send('收到'.encode(self.encoding))
            except Exception as e:
                self.stream.system_out_error(f'client: {username}{address} 断开连接: {e}')
                client.close()
                if username in self.client_pool.keys(): self.client_pool.pop(username)
                self.send_all(f'用户: {username}, 离开聊天室')
                break

    def start(self):
        try:
            self.stream.start_service()
            self.stream.system_out_normal('letter2 2.0 service starting...')
            self.service.bind((self.service_host, self.service_port))
            self.service.listen(self.max_client)
            self.stream.system_out_win('letter2 2.0 service stared!')
            self.stream.system_out_win(f'letter2 2.0 service host: {self.service_host}')
            self.stream.system_out_win(f'letter2 2.0 service port: {self.service_port}')

            while True:
                client, address = self.service.accept()
                order = client.recv(self.buffer_size).decode(self.encoding)
                try:
                    # 登录 / 注册 校验
                    order, username, password = order.split(' ')
                    result = self.login_or_register(order=order, username=username, password=password)
                    client.send(result.encode(self.encoding))
                    if 'err' in result:
                        # 注册或登录失败
                        client.close()
                        continue

                    # 登录 / 注册 成功, 给与日志输出
                    self.stream.system_out_warning(f'client: {username}{address} 加入聊天')
                    self.send_all(f'用户 {username} 加入聊天')

                    # 开启线程为其服务
                    ts = threading.Thread(target=self.main, args=(client, address, username))
                    ts.start()
                except Exception as e:
                    self.stream.system_out_error(f'client {address}, 的错误访问: {e}')
                    client.close()

        except Exception as e:
            self.stream.system_out_error(f'服务器启动异常: {e}')
            self.service.close()


# ======================================================================================================================
class Client(Base):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.username = '离线状态'
        self.client = socket.socket()
        self.address = ''
        self.lock = threading.Lock()

    def login_or_register(self, order) -> bool:
        """ 登录 / 注册 """
        username = self.verify.username_verify()
        password = self.verify.password_verify()
        try:
            self.stream.system_out_normal(f'{order}...')
            self.client.connect((self.service_host, self.service_port))
            self.client.send(f'{order} {username} {password}'.encode(self.encoding))
            result = self.client.recv(self.buffer_size).decode(self.encoding)
            if 'ok' in result:
                self.address = self.client.getsockname()
                self.username = username
                self.stream.system_out_win(f'登录成功, 目标服务器: {self.service_host}:{self.service_port}')
                return True
            else:
                self.stream.system_out_error(result)
                self.client = socket.socket()
        except Exception as e:
            self.stream.system_out_error(f'服务器{self.service_host}:{self.service_port}, 连接失败, 请稍后重试: {e}')

    def off_main(self):
        """ 离线服务 """
        while True:
            order = self.stream.user_input().strip()
            if not order: continue
            if self.verify.order_verify(order=order):
                # 登录 / 注册, 成功的话关闭离线服务, 开启线上服务
                if order == 'login' or order == 'register':
                    if self.login_or_register(order=order):
                        self.main()
                        break

                # 注销账号
                elif order == 'exit':
                    break

                # 其它服务
                elif not self.current(order):
                    self.stream.system_out_warning(f'离线模式无法使用该指令: {order}')
            else:
                self.stream.system_out_warning('错误指令, 你可以输入 help 寻求帮助')

    def monitorUser(self):
        """ 监控客户端用户输入 """
        while True:
            order = input().strip()
            if not order: continue
            if self.verify.order_verify(order=order):
                if order == 'exit':
                    # 退出应用
                    self.client.send('exit'.encode(self.encoding))
                    break

                elif order == 'logout':
                    # 注销登录
                    self.client.send('exit'.encode(self.encoding))
                    time.sleep(0.1)
                    self.username = '离线模式'
                    self.address = ''
                    self.client = socket.socket()
                    self.off_main()
                    break

                elif order.startswith('file'):
                    self.lock.acquire()
                    # 文件传输: 校验文件真实性，告诉服务器文件要发送了，上传文件
                    order, filename, to, user = order.split(' ')
                    flag, filesize = self.file.verify(filename)
                    if not flag:
                        # 校验文件是否存在
                        self.stream.system_out_warning(f'文件不存在！或者目标是个目录！')
                        continue

                    self.client.send(f'{order} {filename} {filesize} to {user}'.encode(self.encoding))  # 告诉服务端，文件来了
                    self.file.upload_file(filename, filesize=filesize, client=self.client)
                    self.lock.release()

                elif order == 'ls' or order.startswith('chat') or order.startswith('repwd'):
                    self.client.send(order.encode(self.encoding))

                elif not self.current(order):
                    self.stream.system_out_warning(f'在线模式无法使用该指令: {order}')
            else:
                self.stream.system_out_warning('错误指令, 你可以输入 help 寻求帮助')

    def monitorService(self):
        """ 监控服务端返回 """
        while True:
            try:
                result = self.client.recv(self.buffer_size)
                if result == self.file.file_over: continue
                result = result.decode(self.encoding)
            except Exception as e:
                continue

            if result == 'exit':
                self.stream.system_out_error('goodbye :)')
                break

            if result.startswith('file'):
                self.lock.acquire()
                order, filename, filesize, old_user = result.split(' ')
                self.file.save(filename=filename, filesize=int(filesize), username=self.username, old_user=old_user,
                               client=self.client)
                self.lock.release()

            else:
                self.stream.system_out_normal(result)

    def main(self):
        """ 线上服务 """
        mu = threading.Thread(target=self.monitorUser)
        ms = threading.Thread(target=self.monitorService, daemon=True)
        mu.start()
        ms.start()
        ms.join()

    def start(self):
        self.stream.start_client()
        self.stream.system_out_normal('letter2 2.0 client starting...')
        self.stream.system_out_win('letter2 2.0 client stared!')
        self.stream.system_out_warning('目前正处于离线模式, 你可以输入 login 登录聊天室')

        # 进入离线服务
        self.off_main()


# 工具包
import os.path
import re
import time
from datetime import datetime
from tqdm import tqdm


class Stream:
    __service_start_log = """
 _      _   _                                  _            ___    ___  
 | |    | | | |                                (_)          |__ \  / _ \ 
 | | ___| |_| |_ ___ _ __   ___  ___ _ ____   ___  ___ ___     ) || | | |
 | |/ _ \ __| __/ _ \ '__| / __|/ _ \ '__\ \ / / |/ __/ _ \   / / | | | |
 | |  __/ |_| ||  __/ |    \__ \  __/ |   \ V /| | (_|  __/  / /_ | |_| |
 |_|\___|\__|\__\___|_|    |___/\___|_|    \_/ |_|\___\___| |____(_)___/                                                                                                           
"""
    __client_start_log = """
  _      _   _                   _ _            _     ___    ___  
 | |    | | | |                 | (_)          | |   |__ \  / _ \ 
 | | ___| |_| |_ ___ _ __    ___| |_  ___ _ __ | |_     ) || | | |
 | |/ _ \ __| __/ _ \ '__|  / __| | |/ _ \ '_ \| __|   / / | | | |
 | |  __/ |_| ||  __/ |    | (__| | |  __/ | | | |_   / /_ | |_| |
 |_|\___|\__|\__\___|_|     \___|_|_|\___|_| |_|\__| |____(_)___/ 
"""
    __help_text = """
    绿色表示\033[0;32m在线\033[0m才能使用的功能
    红色表示\033[0;31m离线\033[0m才能使用的功能
    白色不做限制
    
        \033[0;32mls\033[0m: 显示所有在线用户
        me: 显示自己的状态
        clear: 清空记录
        help: 查看帮助
        path: 显示当前目录，及目录下的所有子目录
        \033[0;31mlogin\033[0m: 登录服务器
        \033[0;31mregister\033[0m: 注册一个账户并登录服务器
        \033[0;32mchat\033[0m [all|username] [text]: 给[所有|指定]用户发送消息
        \033[0;32mfile\033[0m [filepath] to [username]: 给指定用户发送文件
        \033[0;32mlogout\033[0m: 退出登录
        \033[0;32mrepwd\033[0m [old password] [new password]: 修改密码
        exit: 退出应用
    """

    @staticmethod
    def __now_date() -> str:
        return datetime.now().strftime('%Y-%m-%d %H:%M:%S')

    def __system_out(self, msg: str, date=True):
        msg = f'$ {self.__now_date()} > {msg}' if date else f'$ > {msg}'
        print(msg)

    def system_out_normal(self, msg: str, date=True):
        """ 标准系统输出 """
        message = f'\033[0;36m{msg}\033[0m'
        self.__system_out(msg=message, date=date)

    def system_out_warning(self, msg: str, date=True):
        """ 系统警告输出 """
        message = f'\033[0;33m{msg}\033[0m'
        self.__system_out(msg=message, date=date)

    def system_out_error(self, msg: str, date=True):
        """ 系统错误输出 """
        message = f'\033[0;31m{msg}\033[0m'
        self.__system_out(msg=message, date=date)

    def system_out_win(self, msg: str, date=True):
        """ 系统成功输出 """
        message = f'\033[0;32m{msg}\033[0m'
        self.__system_out(msg=message, date=date)

    def system_out_help(self):
        print(self.__help_text)

    def start_service(self):
        print(self.__service_start_log)

    def start_client(self):
        print(self.__client_start_log)

    def system_out_path(self):
        bas_dir = os.getcwd()
        list_dir = os.listdir(bas_dir)
        print(f'\n  \033[0;36m{bas_dir}\033[0m')
        print(f'\tpath type\tpath size\tpath name')
        for name in list_dir:
            if os.path.isdir(f'{bas_dir}/{name}'):
                print('\tdir\t', end='\t')
            else:
                print('\tfile\t', end='\t')

            print('{0:>}'.format(os.path.getsize(f'{bas_dir}/{name}')), end='\t\t')
            print('{0:>}'.format(name))



    @staticmethod
    def system_input(msg: str) -> str:
        return input(f'$ {msg}: ')

    @staticmethod
    def user_input() -> str:
        return input(f'# ')


class Verify:
    @staticmethod
    def username_verify() -> str:
        username = Stream.system_input("username").strip()
        while True:
            if username.find(' ') != -1:
                username = Stream.system_input("不能包含空格, username")
            elif username in ['all', 'exit', 'file', 'to', 'chat', 'clear', 'me']:
                username = Stream.system_input("不能是关键字, username")
            elif not (4 <= len(username) <= 8):
                username = Stream.system_input("长度必须是4~8, username")
            else:
                break
        return username

    @staticmethod
    def password_verify() -> str:
        password = Stream.system_input("password").strip()
        while True:
            if re.match('^[a-z0-9]{4,8}$', password): break
            password = Stream.system_input("密码由数组和字母组成的4~8位, password").strip()
        return password

    @staticmethod
    def order_verify(order: str) -> bool:
        """ 校验指令 """
        if order == 'ls' or order == 'me': return True
        if order == 'clear' or order == 'login': return True
        if order == 'register' or order == 'logout': return True
        if order == 'exit' or order == 'help' or order == 'path': return True
        if order.startswith('chat') and len(order.split(' ')) >= 3: return True
        if order.startswith('file') and len(order.split(' ')) == 4 and order.split()[-2] == 'to': return True
        if order.startswith('repwd') and len(order.split(' ')) == 3: return True
        return False


class File:
    file_speed: int = 2048
    file_over: str = b'file over bytes'

    @staticmethod
    def verify(filename: str) -> tuple:
        """ 校验文件并返回文件大小 """
        if os.path.exists(filename): return True, os.path.getsize(filename)
        return False, 0

    def upload_file(self, filename: str, filesize, client):
        progress = tqdm(range(filesize), f'发送: {filename}', unit='B', unit_divisor=1024, unit_scale=True)
        with open(filename, 'rb') as fp:
            while True:
                time.sleep(0.1)
                data = fp.read(self.file_speed)
                if not data:
                    client.send(self.file_over)
                    break
                client.sendall(data)
                progress.update(len(data))

    def download_file(self, client, filesize, filename):
        progress = tqdm(range(filesize), f'接收: {filename}', unit='B', unit_divisor=1024, unit_scale=True)
        if not os.path.exists(f'./temp'): os.mkdir(f'./temp')
        with open(f'./temp/{filename}', 'wb') as fp:
            while True:
                item = client.recv(self.file_speed)
                if item == self.file_over:
                    client.send(self.file_over)
                    break
                fp.write(item)
                progress.update(len(item))

    def save(self, filename, filesize, username, old_user, client):
        if not os.path.exists(f'./{username}'): os.mkdir(f'./{username}')
        filepath = f'./{username}/{filename}'
        progress = tqdm(range(filesize), f'接收: {filename}', unit='B', unit_divisor=1024, unit_scale=True)
        with open(filepath, 'wb') as fp:
            while True:
                data = client.recv(self.file_speed)
                if data == self.file_over:
                    break
                fp.write(data)
                progress.update(len(data))

        Stream().system_out_win(f'用户 {old_user} 传递的文件 {filename} {filesize}B, 保存至: {filepath}')



def service(**kwargs):
    """
    返回一个服务器对象
    # 数据库配置【以下都是默认值】
    mysql_host='localhost',
    mysql_port=3306,
    mysql_user='root',
    mysql_database='letter',
    mysql_password='1234',

    # 服务器端配置【以下都是默认值】
    service_host='localhost',
    service_port=5000,
    max_client=5,
    encoding='utf8'
    """
    from .core import Service
    return Service(**kwargs)


def client(**kwargs):
    """
    返回一个客服端对象
    service_host: 服务端地址, default: localhost
    service_port: 服务端端口, default: 5000
    encoding: 编码方式， default: utf-8
    """
    from .core import Client
    return Client(**kwargs)



from .utils import Stream
import pymysql


class DataBaseManager:
    def __init__(self, **kwargs):
        self.mysql_host = kwargs.get('mysql_host', '127.0.0.1')
        self.mysql_port = kwargs.get('mysql_port', 3306)
        self.mysql_username = kwargs.get('mysql_username', 'root')
        self.mysql_password = kwargs.get('mysql_password', '1234')
        self.mysql_database = kwargs.get('mysql_database', 'letter')
        self.stream = Stream()
        self.connection = pymysql.connect(host=self.mysql_host, port=self.mysql_port,
                                          user=self.mysql_username, password=self.mysql_password,
                                          database=self.mysql_database)

        self.init_database()

    def login(self, username: str, password: str) -> bool:
        cursor = self.connection.cursor()
        cursor.execute('select * from letter_user where username = "%s"' % username)
        user = cursor.fetchall()
        if user:
            # 用户存在
            cursor.close()
            return user[0][2] == password

        cursor.close()
        return False

    def register(self, username: str, password: str) -> bool:
        cursor = self.connection.cursor()
        result = cursor.execute('select * from letter_user where username = "%s"' % username)
        if result:
            # 如果用户已经存在
            cursor.close()
            return False

        cursor.execute('insert into letter_user(username, password) values("%s", "%s")' % (username, password))
        self.connection.commit()
        cursor.close()
        return True

    def re_pwd(self, username: str, old_pwd: str, new_pwd: str) -> bool:
        cursor = self.connection.cursor()
        cursor.execute('select * from letter_user where username = "%s"' % username)
        user = cursor.fetchall()
        if user and user[0][2] == old_pwd:
            # 如果用户存在且密码匹配
            cursor.execute('update letter_user set password = "%s" where username = "%s"' % (new_pwd, username))
            self.connection.commit()
            cursor.close()
            return True

        cursor.close()
        return False

    def init_database(self):
        cursor = self.connection.cursor()
        result = cursor.execute(f'show tables where Tables_in_{self.mysql_database} = "letter_user"')
        if result == 0:
            self.stream.system_out_normal('正在初始化数据库中！')
            cursor.execute("""
                create table if not exists letter_user(
                    id int primary key auto_increment,
                    username varchar(20) not null unique,
                    password varchar(32)
                )
            """)
            self.stream.system_out_normal('表 letter_user 构建完毕')
        cursor.close()
